from typing import Union, TypeVar

K = TypeVar('K')
V = TypeVar('V')


class FormattedBase:
    """Base class for any message type."""
    text: str

    def __add__(self, other: Union[V, 'FormattedBase']) -> str:
        return str(self) + str(other)

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self.text})'

    def __str__(self) -> str:
        return self.text


class Bold(FormattedBase):
    def __init__(self, text: V) -> None:
        """A bold text

        Args:
            text: Text
        """
        self.text = f'**{text}**'


class Italic(FormattedBase):
    def __init__(self, text: V) -> None:
        """A italic text

        Args:
            text: Text
        """
        self.text = f'__{text}__'


class Strikethrough(FormattedBase):
    def __init__(self, text: V) -> None:
        """A strikethrough text

        Args:
            text: Text
        """
        self.text = f'~~{text}~~'


class Code(FormattedBase):
    def __init__(self, text: V) -> None:
        """Text formatted as code (monospace)

        Args:
            text: Text
        """
        self.text = f'`{text}`'


class Pre(FormattedBase):
    def __init__(self, text: V) -> None:
        """A Multiline Monospaced text.

        Args:
            text: Text
        """
        self.text = f'```{text}```'


class Link(FormattedBase):
    def __init__(self, label: V, url: str) -> None:
        """A Hyperlink with a label.

        Args:
            label: Label of the link
            url: URL of the link
        """
        self.text = f'[{label}]({url})'



from ._document import KanTeXDocument
from ._items import Mention, KeyValueItem, Item
from ._sections import Section, SubSection, SubSubSection
from ._styles import Bold, Italic, Strikethrough, Code, Pre, Link



from typing import Union, TypeVar, Tuple, Type

from ._styles import Link, FormattedBase

K = TypeVar('K')
V = TypeVar('V')


class Mention(Link):
    def __init__(self, label: V, uid: int):
        """Inline Mention of a User.

        Args:
            label: The text of the mention
            uid: The id of the user to be mentioned
        """
        super().__init__(label, f'tg://user?id={uid}')


class KeyValueItem(FormattedBase):
    def __init__(self, key: Union[K, FormattedBase], value: Union[V, FormattedBase],
                 colon_styles: Tuple[Type[FormattedBase]] = None) -> None:
        """A item that has a key and a value divided by a colon.

        Args:
            key: The key
            value: The Value
            colon_styles: Tuple of style to be used for the colon
        """
        self.key = key
        self.value = value
        colon = ':'
        if colon_styles is not None:
            for style in colon_styles:
                colon = style(colon)

        self.text = f'{key}{colon} {value}'


class Item(FormattedBase):
    def __init__(self, text: V) -> None:
        """A simple item without any formatting.

        Args:
            text: The text
        """
        self.text = str(text)



from typing import Union, Iterable, TypeVar

K = TypeVar('K')
V = TypeVar('V')


class KanTeXDocument:
    def __init__(self, *args: Union[V, 'Section']) -> None:
        """Container class for Sections and Items

        Args:
            *args: Initial items of the document
        """
        self.sections = [i for i in args if i]

    def __str__(self) -> str:
        return '\n\n'.join([str(section) for section in self.sections])

    def append(self, item: Union[V, 'FormattedBase', 'Section']) -> None:
        """Append a section or item to the document

        Args:
            item: The item to append
        """
        self.sections.append(item)

    def extend(self, items: Iterable) -> None:
        """Extend the document with a list of items

        Args:
            items: List of items to extend the document with
        """
        self.sections.extend(items)



from typing import Union, Iterable, TypeVar

from ._document import KanTeXDocument
from ._styles import FormattedBase, Bold

K = TypeVar('K')
V = TypeVar('V')


class Section:
    def __init__(self, *args: Union[V, 'SubSection'], indent: int = 4) -> None:
        """A section which automatically indents its contents

        The header is made bold automatically

        Args:
            *args: Initial items
            indent: Ident to use
        """
        self.header = Bold(args[0])
        self.items = [i for i in args[1:] if i]
        self.indent = indent

    def __add__(self, other: Union[V, 'Section']) -> 'MDTeXDocument':
        return KanTeXDocument(self, *other)

    def __str__(self) -> str:
        return '\n'.join([str(self.header)]
                         + [' ' * self.indent + str(item) for item in self.items
                            if item is not None])

    def append(self, item: Union[V, FormattedBase, 'SubSection']) -> None:
        """Append a subsection or item to the section

        Args:
            item: The item to append
        """
        self.items.append(item)

    def extend(self, items: Iterable) -> None:
        """Extend the section with a list of items

        Args:
            items: List of items to extend the document with
        """
        self.items.extend(items)


class SubSection(Section):
    def __init__(self, *args: Union[V, 'SubSubSection'], indent: int = 8) -> None:
        """A sub-section which automatically indents its contents

        The header is made bold automatically

        Args:
            *args: Initial items
            indent: Ident to use
        """
        super().__init__(*args, indent=indent)


class SubSubSection(SubSection):
    def __init__(self, *args: V, indent: int = 12) -> None:
        """A sub-sub-section which automatically indents its contents

        The header is made bold automatically

        Args:
            *args: Initial items
            indent: Ident to use
        """
        super().__init__(*args, indent=indent)
